MODULE asmbgc
   !!===========================================================================
   !!                       ***  MODULE asmbgc  ***
   !! Routines and declarations for biogeochemical assimilation
   !!===========================================================================
   !! History :  3.6  ! 2018-02 (D. Ford)  Adapted from existing modules
   !!---------------------------------------------------------------------------
   !! 'key_asminc'          : assimilation increment interface
   !! 'key_top'             : passive tracers
   !! 'key_hadocc'          : HadOCC model
   !! 'key_medusa'          : MEDUSA model
   !! 'key_foam_medusa'     : MEDUSA extras for FOAM OBS and ASM
   !! 'key_roam'            : MEDUSA extras for carbonate cycle
   !! 'key_karaml'          : Kara mixed layer depth
   !!---------------------------------------------------------------------------
   !! asm_bgc_check_options : check bgc assimilation options
   !! asm_bgc_init_incs     : initialise bgc increments
   !! asm_bgc_init_bkg      : initialise bgc background
   !! asm_bgc_bal_wri       : write out bgc balancing increments
   !! asm_bgc_bkg_wri       : write out bgc background
   !! phyto2d_asm_inc       : apply the ocean colour increments
   !! phyto3d_asm_inc       : apply the 3D phytoplankton increments
   !! pco2_asm_inc          : apply the pCO2/fCO2 increments
   !! ph_asm_inc            : apply the pH increments
   !! bgc3d_asm_inc         : apply the generic 3D BGC increments
   !!---------------------------------------------------------------------------
   USE par_kind, ONLY:      & ! kind parameters
      & wp
   USE par_oce, ONLY:       & ! domain array sizes
      & jpi,                &
      & jpj,                &
      & jpk
   USE iom                    ! i/o
   USE oce, ONLY:           & ! active tracer variables
      & tsn
   USE zdfmxl, ONLY :       & ! mixed layer depth 
#if defined key_karaml
      & hmld_kara,          &
      & ln_kara,            &
#endif   
      & hmld,               & 
      & hmlp,               &
      & hmlpt 
   USE asmpar, ONLY:        & ! assimilation parameters
      & c_asmbkg,           &
      & c_asmbal,           &
      & nitbkg_r,           &
      & nitdin_r,           &
      & nitiaustr_r,        &
      & nitiaufin_r
#if defined key_top
   USE trc, ONLY:           & ! passive tracer variables
      & trn,                &
      & trb
   USE par_trc, ONLY:       & ! passive tracer parameters
      & jptra
#endif
#if defined key_medusa && defined key_foam_medusa
   USE asmlogchlbal_medusa, ONLY: & ! logchl balancing for MEDUSA
      & asm_logchl_bal_medusa
   USE asmpco2bal, ONLY:    & ! pCO2 balancing for MEDUSA
      & asm_pco2_bal
   USE par_medusa             ! MEDUSA parameters
   USE mocsy_mainmod, ONLY: & ! MEDUSA carbonate system
      & f2pCO2
   USE sms_medusa, ONLY:    & ! MEDUSA diagnostic variables
      & pgrow_avg,          &
      & ploss_avg,          &
      & phyt_avg,           &
      & mld_max
#elif defined key_hadocc
   USE asmlogchlbal_hadocc, ONLY: & ! logchl balancing for HadOCC
      & asm_logchl_bal_hadocc
   USE asmpco2bal, ONLY:    & ! pCO2 balancing for HadOCC
      & asm_pco2_bal
   USE par_hadocc             ! HadOCC parameters
   USE had_bgc_const          ! HadOCC parameters
   USE trc, ONLY:           & ! HadOCC diagnostic variables
      & pgrow_avg,          &
      & ploss_avg,          &
      & phyt_avg,           &
      & mld_max,            &
      & HADOCC_CHL
   USE had_bgc_const, ONLY: & ! HadOCC C:Chl ratio
      & cchl_p
#endif

   IMPLICIT NONE
   PRIVATE                   

   PUBLIC  asm_bgc_check_options  ! called by asm_inc_init in asminc.F90
   PUBLIC  asm_bgc_init_incs      ! called by asm_inc_init in asminc.F90
   PRIVATE asm_bgc_read_incs_2d   ! called by asm_bgc_init_incs
   PRIVATE asm_bgc_read_incs_3d   ! called by asm_bgc_init_incs
   PUBLIC  asm_bgc_init_bkg       ! called by asm_inc_init in asminc.F90
   PUBLIC  asm_bgc_bal_wri        ! called by nemo_gcm in nemogcm.F90
   PUBLIC  asm_bgc_bkg_wri        ! called by asm_bkg_wri in asmbkg.F90
   PUBLIC  phyto2d_asm_inc        ! called by bgc_asm_inc in asminc.F90
   PUBLIC  phyto3d_asm_inc        ! called by bgc_asm_inc in asminc.F90
   PUBLIC  pco2_asm_inc           ! called by bgc_asm_inc in asminc.F90
   PUBLIC  ph_asm_inc             ! called by bgc_asm_inc in asminc.F90
   PUBLIC  bgc3d_asm_inc          ! called by bgc_asm_inc in asminc.F90

   LOGICAL, PUBLIC :: ln_balwri      = .FALSE. !: No output of balancing incs
   LOGICAL, PUBLIC :: ln_phytobal    = .FALSE. !: No phytoplankton balancing
   LOGICAL, PUBLIC :: ln_slchltotinc = .FALSE. !: No surface total      log10(chlorophyll) increment
   LOGICAL, PUBLIC :: ln_slchldiainc = .FALSE. !: No surface diatom     log10(chlorophyll) increment
   LOGICAL, PUBLIC :: ln_slchlnoninc = .FALSE. !: No surface non-diatom log10(chlorophyll) increment
   LOGICAL, PUBLIC :: ln_schltotinc  = .FALSE. !: No surface total      chlorophyll        increment
   LOGICAL, PUBLIC :: ln_slphytotinc = .FALSE. !: No surface total      log10(phyto C)     increment
   LOGICAL, PUBLIC :: ln_slphydiainc = .FALSE. !: No surface diatom     log10(phyto C)     increment
   LOGICAL, PUBLIC :: ln_slphynoninc = .FALSE. !: No surface non-diatom log10(phyto C)     increment
   LOGICAL, PUBLIC :: ln_spco2inc    = .FALSE. !: No surface pCO2                          increment
   LOGICAL, PUBLIC :: ln_sfco2inc    = .FALSE. !: No surface fCO2                          increment
   LOGICAL, PUBLIC :: ln_plchltotinc = .FALSE. !: No profile total      log10(chlorophyll) increment
   LOGICAL, PUBLIC :: ln_pchltotinc  = .FALSE. !: No profile total      chlorophyll        increment
   LOGICAL, PUBLIC :: ln_pno3inc     = .FALSE. !: No profile nitrate                       increment
   LOGICAL, PUBLIC :: ln_psi4inc     = .FALSE. !: No profile silicate                      increment
   LOGICAL, PUBLIC :: ln_pdicinc     = .FALSE. !: No profile dissolved inorganic carbon    increment
   LOGICAL, PUBLIC :: ln_palkinc     = .FALSE. !: No profile alkalinity                    increment
   LOGICAL, PUBLIC :: ln_pphinc      = .FALSE. !: No profile pH                            increment
   LOGICAL, PUBLIC :: ln_po2inc      = .FALSE. !: No profile oxygen                        increment

   INTEGER, PUBLIC :: mld_choice_bgc = 1 !: choice of mld for bgc assimilation
                                         !  1) hmld      - Turbocline/mixing depth
                                         !                 [W points]
                                         !  2) hmlp      - Density criterion
                                         !                 (0.01 kg/m^3 change from 10m)
                                         !                 [W points]
                                         !  3) hmld_kara - Kara MLD
                                         !                 [Interpolated]
                                         !  4) hmld_tref - Temperature criterion
                                         !                 (0.2 K change from surface)
                                         !                 [T points]
                                         !  5) hmlpt     - Density criterion
                                         !                 (0.01 kg/m^3 change from 10m)
                                         !                 [T points]

   REAL(wp), PUBLIC :: rn_maxchlinc = -999.0 !: maximum absolute non-log chlorophyll inc
                                             ! <= 0 no maximum applied (switch turned off)
                                             !  > 0 absolute chl inc capped at this value

   REAL(wp), DIMENSION(:,:),     ALLOCATABLE :: slchltot_bkginc ! slchltot inc
   REAL(wp), DIMENSION(:,:),     ALLOCATABLE :: slchldia_bkginc ! slchldia inc
   REAL(wp), DIMENSION(:,:),     ALLOCATABLE :: slchlnon_bkginc ! slchlnon inc
   REAL(wp), DIMENSION(:,:),     ALLOCATABLE :: schltot_bkginc  ! schltot inc
   REAL(wp), DIMENSION(:,:),     ALLOCATABLE :: slphytot_bkginc ! slphytot inc
   REAL(wp), DIMENSION(:,:),     ALLOCATABLE :: slphydia_bkginc ! slphydia inc
   REAL(wp), DIMENSION(:,:),     ALLOCATABLE :: slphynon_bkginc ! slphynon inc
   REAL(wp), DIMENSION(:,:),     ALLOCATABLE :: sfco2_bkginc    ! sfco2 inc
   REAL(wp), DIMENSION(:,:),     ALLOCATABLE :: spco2_bkginc    ! spco2 inc
   REAL(wp), DIMENSION(:,:,:),   ALLOCATABLE :: plchltot_bkginc ! plchltot inc
   REAL(wp), DIMENSION(:,:,:),   ALLOCATABLE :: pchltot_bkginc  ! pchltot inc
   REAL(wp), DIMENSION(:,:,:),   ALLOCATABLE :: pno3_bkginc     ! pno3 inc
   REAL(wp), DIMENSION(:,:,:),   ALLOCATABLE :: psi4_bkginc     ! psi4 inc
   REAL(wp), DIMENSION(:,:,:),   ALLOCATABLE :: pdic_bkginc     ! pdic inc
   REAL(wp), DIMENSION(:,:,:),   ALLOCATABLE :: palk_bkginc     ! palk inc
   REAL(wp), DIMENSION(:,:,:),   ALLOCATABLE :: pph_bkginc      ! pph inc
   REAL(wp), DIMENSION(:,:,:),   ALLOCATABLE :: po2_bkginc      ! po2 inc
#if defined key_top
   REAL(wp), DIMENSION(:,:,:,:), ALLOCATABLE :: phyto2d_balinc  ! Balancing incs from ocean colour
   REAL(wp), DIMENSION(:,:,:,:), ALLOCATABLE :: phyto3d_balinc  ! Balancing incs from p(l)chltot
   REAL(wp), DIMENSION(:,:,:,:), ALLOCATABLE :: pco2_balinc     ! Balancing incs from spco2/sfco2
   REAL(wp), DIMENSION(:,:,:,:), ALLOCATABLE :: ph_balinc       ! Balancing incs from pph
#endif

#if defined key_hadocc || (defined key_medusa && defined key_foam_medusa)
   REAL(wp), DIMENSION(:,:),     ALLOCATABLE :: pgrow_avg_bkg   ! Background phyto growth
   REAL(wp), DIMENSION(:,:),     ALLOCATABLE :: ploss_avg_bkg   ! Background phyto loss
   REAL(wp), DIMENSION(:,:),     ALLOCATABLE :: phyt_avg_bkg    ! Background phyto conc
   REAL(wp), DIMENSION(:,:),     ALLOCATABLE :: mld_max_bkg     ! Background max MLD
   REAL(wp), DIMENSION(:,:,:,:), ALLOCATABLE :: tracer_bkg      ! Background tracer state
#endif
#if defined key_hadocc
   REAL(wp), DIMENSION(:,:,:),   ALLOCATABLE :: chl_bkg         ! Background chl
   REAL(wp), DIMENSION(:,:,:),   ALLOCATABLE :: cchl_p_bkg      ! Background c:chl
#endif

CONTAINS

   SUBROUTINE asm_bgc_check_options
      !!------------------------------------------------------------------------
      !!                    ***  ROUTINE asm_bgc_check_options  ***
      !!
      !! ** Purpose :   check validity of biogeochemical assimilation options
      !!
      !! ** Method  :   check settings of logicals
      !!
      !! ** Action  :   call ctl_stop/ctl_warn if required
      !!
      !! References :   asm_inc_init
      !!------------------------------------------------------------------------

#if ! defined key_top || ( ! defined key_hadocc && ( ! defined key_medusa || ! defined key_foam_medusa ) )
      CALL ctl_stop( ' Attempting to assimilate biogeochemical observations', &
         &           ' but no compatible biogeochemical model is available' )
#endif

#if defined key_hadocc
      IF ( ln_slchldiainc .OR. ln_slchlnoninc .OR. ln_slphydiainc .OR. &
         & ln_slphynoninc .OR. ln_psi4inc .OR. ln_pphinc .OR. ln_po2inc ) THEN
         CALL ctl_stop( ' Cannot assimilate PFTs, Si4, pH or O2 into HadOCC' )
      ENDIF
#endif

      IF ( ( ln_phytobal ).AND.                                       &
         & ( .NOT. ln_slchltotinc ).AND.( .NOT. ln_slchldiainc ).AND. &
         & ( .NOT. ln_slchlnoninc ).AND.( .NOT. ln_schltotinc  ).AND. &
         & ( .NOT. ln_slphytotinc ).AND.( .NOT. ln_slphydiainc ).AND. &
         & ( .NOT. ln_slphynoninc ) ) THEN
         CALL ctl_warn( ' Cannot calculate phytoplankton balancing increments', &
            &           ' if not assimilating ocean colour,',                   &
            &           ' so ln_phytobal will be set to .false.')
         ln_phytobal = .FALSE.
      ENDIF

      IF ( ( ln_balwri ).AND.( .NOT. ln_phytobal ).AND. &
         & ( .NOT. ln_spco2inc ).AND.( .NOT. ln_sfco2inc ).AND.( .NOT. ln_pphinc ) ) THEN
         CALL ctl_warn( ' Balancing increments are only calculated for ocean colour', &
            &           ' with ln_phytobal, or for pCO2, fCO2, or pH.',               &
            &           ' Not the case, so ln_balwri will be set to .false.')
         ln_balwri = .FALSE.
      ENDIF

      IF ( ln_spco2inc .AND. ln_sfco2inc ) THEN
         CALL ctl_stop( ' Can only assimilate pCO2 OR fCO2, not both' )
      ENDIF

      IF ( ln_slchltotinc .AND. ln_schltotinc ) THEN
         CALL ctl_stop( ' Can only assimilate surface log10(chlorophyll) or chlorophyll, not both' )
      ENDIF

      IF ( ln_plchltotinc .AND. ln_pchltotinc ) THEN
         CALL ctl_stop( ' Can only assimilate profile log10(chlorophyll) or chlorophyll, not both' )
      ENDIF

      IF ( ( ln_slchltotinc .OR. ln_schltotinc  ) .AND. &
         & ( ln_slchldiainc .OR. ln_slchlnoninc ) ) THEN
         CALL ctl_stop( ' Can only assimilate total or PFT surface chlorophyll, not both' )
      ENDIF

      IF ( ln_slphytotinc .AND. ( ln_slphydiainc .OR. ln_slphynoninc ) ) THEN
         CALL ctl_stop( ' Can only assimilate total or PFT surface phytoplankton carbon, not both' )
      ENDIF

   END SUBROUTINE asm_bgc_check_options

   !!===========================================================================
   !!===========================================================================
   !!===========================================================================

   SUBROUTINE asm_bgc_init_incs( knum )
      !!------------------------------------------------------------------------
      !!                    ***  ROUTINE asm_bgc_init_incs  ***
      !!
      !! ** Purpose :   initialise bgc increments
      !!
      !! ** Method  :   allocate arrays and read increments from file
      !!
      !! ** Action  :   allocate arrays and read increments from file
      !!
      !! References :   asm_inc_init
      !!------------------------------------------------------------------------
      !!
      INTEGER, INTENT(in   ) :: knum   ! i/o unit of increments file
      !!
      !!------------------------------------------------------------------------

      ! Allocate and read increments
      
      IF ( ln_slchltotinc ) THEN
         ALLOCATE( slchltot_bkginc(jpi,jpj) )
         CALL asm_bgc_read_incs_2d( knum, 'bckinslchltot', slchltot_bkginc )
      ENDIF
      
      IF ( ln_slchldiainc ) THEN
         ALLOCATE( slchldia_bkginc(jpi,jpj) )
         CALL asm_bgc_read_incs_2d( knum, 'bckinslchldia', slchldia_bkginc )
      ENDIF
      
      IF ( ln_slchlnoninc ) THEN
         ALLOCATE( slchlnon_bkginc(jpi,jpj) )
         CALL asm_bgc_read_incs_2d( knum, 'bckinslchlnon', slchlnon_bkginc )
      ENDIF
      
      IF ( ln_schltotinc ) THEN
         ALLOCATE( schltot_bkginc(jpi,jpj) )
         CALL asm_bgc_read_incs_2d( knum, 'bckinschltot', schltot_bkginc )
      ENDIF
      
      IF ( ln_slphytotinc ) THEN
         ALLOCATE( slphytot_bkginc(jpi,jpj) )
         CALL asm_bgc_read_incs_2d( knum, 'bckinslphytot', slphytot_bkginc )
      ENDIF
      
      IF ( ln_slphydiainc ) THEN
         ALLOCATE( slphydia_bkginc(jpi,jpj) )
         CALL asm_bgc_read_incs_2d( knum, 'bckinslphydia', slphydia_bkginc )
      ENDIF
      
      IF ( ln_slphynoninc ) THEN
         ALLOCATE( slphynon_bkginc(jpi,jpj) )
         CALL asm_bgc_read_incs_2d( knum, 'bckinslphynon', slphynon_bkginc )
      ENDIF

      IF ( ln_sfco2inc ) THEN
         ALLOCATE( sfco2_bkginc(jpi,jpj) )
         CALL asm_bgc_read_incs_2d( knum, 'bckinsfco2', sfco2_bkginc )
      ENDIF

      IF ( ln_spco2inc ) THEN
         ALLOCATE( spco2_bkginc(jpi,jpj) )
         CALL asm_bgc_read_incs_2d( knum, 'bckinspco2', sfco2_bkginc )
      ENDIF
      
      IF ( ln_plchltotinc ) THEN
         ALLOCATE( plchltot_bkginc(jpi,jpj,jpk) )
         CALL asm_bgc_read_incs_3d( knum, 'bckinplchltot', plchltot_bkginc )
      ENDIF
      
      IF ( ln_pchltotinc ) THEN
         ALLOCATE( pchltot_bkginc(jpi,jpj,jpk) )
         CALL asm_bgc_read_incs_3d( knum, 'bckinpchltot', pchltot_bkginc )
      ENDIF
      
      IF ( ln_pno3inc ) THEN
         ALLOCATE( pno3_bkginc(jpi,jpj,jpk) )
         CALL asm_bgc_read_incs_3d( knum, 'bckinpno3', pno3_bkginc )
      ENDIF
      
      IF ( ln_psi4inc ) THEN
         ALLOCATE( psi4_bkginc(jpi,jpj,jpk) )
         CALL asm_bgc_read_incs_3d( knum, 'bckinpsi4', psi4_bkginc )
      ENDIF
      
      IF ( ln_pdicinc ) THEN
         ALLOCATE( pdic_bkginc(jpi,jpj,jpk) )
         CALL asm_bgc_read_incs_3d( knum, 'bckinpdic', pdic_bkginc )
      ENDIF
      
      IF ( ln_palkinc ) THEN
         ALLOCATE( palk_bkginc(jpi,jpj,jpk) )
         CALL asm_bgc_read_incs_3d( knum, 'bckinpalk', palk_bkginc )
      ENDIF
      
      IF ( ln_pphinc ) THEN
         ALLOCATE( pph_bkginc(jpi,jpj,jpk) )
         CALL asm_bgc_read_incs_3d( knum, 'bckinpph', pph_bkginc )
      ENDIF
      
      IF ( ln_po2inc ) THEN
         ALLOCATE( po2_bkginc(jpi,jpj,jpk) )
         CALL asm_bgc_read_incs_3d( knum, 'bckinpo2', po2_bkginc )
      ENDIF

      ! Allocate balancing increments
      
      IF ( ln_slchltotinc .OR. ln_slchldiainc .OR. ln_slchlnoninc .OR. &
         & ln_schltotinc  .OR. ln_slphytotinc .OR. ln_slphydiainc .OR. &
         & ln_slphynoninc ) THEN
#if defined key_top
         ALLOCATE( phyto2d_balinc(jpi,jpj,jpk,jptra) )
         phyto2d_balinc(:,:,:,:) = 0.0
#else
         CALL ctl_stop( ' key_top must be set for balancing increments' )
#endif
      ENDIF
      
      IF ( ln_plchltotinc .OR. ln_pchltotinc ) THEN
#if defined key_top
         ALLOCATE( phyto3d_balinc(jpi,jpj,jpk,jptra) )
         phyto3d_balinc(:,:,:,:) = 0.0
#else
         CALL ctl_stop( ' key_top must be set for balancing increments' )
#endif
      ENDIF

      IF ( ln_spco2inc .OR. ln_sfco2inc ) THEN
#if defined key_top
         ALLOCATE( pco2_balinc(jpi,jpj,jpk,jptra) )
         pco2_balinc(:,:,:,:) = 0.0
#else
         CALL ctl_stop( ' key_top must be set for balancing increments' )
#endif
      ENDIF

      IF ( ln_pphinc ) THEN
#if defined key_top
         ALLOCATE( ph_balinc(jpi,jpj,jpk,jptra) )
         ph_balinc(:,:,:,:) = 0.0
#else
         CALL ctl_stop( ' key_top must be set for balancing increments' )
#endif
      ENDIF

   END SUBROUTINE asm_bgc_init_incs

   !!===========================================================================
   !!===========================================================================
   !!===========================================================================

   SUBROUTINE asm_bgc_read_incs_2d( knum, cd_bgcname, p_incs )
      !!------------------------------------------------------------------------
      !!                    ***  ROUTINE asm_bgc_init_incs  ***
      !!
      !! ** Purpose :   read 2d bgc increments
      !!
      !! ** Method  :   read increments from file
      !!
      !! ** Action  :   read increments from file
      !!
      !! References :   asm_inc_init
      !!------------------------------------------------------------------------
      !!
      INTEGER,                      INTENT(in   ) :: knum       ! i/o unit
      CHARACTER(LEN=13),            INTENT(in   ) :: cd_bgcname ! variable
      REAL(wp), DIMENSION(jpi,jpj), INTENT(  out) :: p_incs     ! increments
      !!
      !!------------------------------------------------------------------------

      ! Initialise
      p_incs(:,:) = 0.0
      
      ! read from file
      CALL iom_get( knum, jpdom_autoglo, TRIM(cd_bgcname), p_incs(:,:), 1 )
      
      ! Apply the masks
      p_incs(:,:) = p_incs(:,:) * tmask(:,:,1)
      
      ! Set missing increments to 0.0 rather than 1e+20
      ! to allow for differences in masks
      WHERE( ABS( p_incs(:,:) ) > 1.0e+10 ) p_incs(:,:) = 0.0

   END SUBROUTINE asm_bgc_read_incs_2d

   !!===========================================================================
   !!===========================================================================
   !!===========================================================================

   SUBROUTINE asm_bgc_read_incs_3d( knum, cd_bgcname, p_incs )
      !!------------------------------------------------------------------------
      !!                    ***  ROUTINE asm_bgc_init_incs  ***
      !!
      !! ** Purpose :   read 3d bgc increments
      !!
      !! ** Method  :   read increments from file
      !!
      !! ** Action  :   read increments from file
      !!
      !! References :   asm_inc_init
      !!------------------------------------------------------------------------
      !!
      INTEGER,                          INTENT(in   ) :: knum       ! i/o unit
      CHARACTER(LEN=13),                INTENT(in   ) :: cd_bgcname ! variable
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT(  out) :: p_incs     ! increments
      !!
      !!------------------------------------------------------------------------

      ! Initialise
      p_incs(:,:,:) = 0.0
      
      ! read from file
      CALL iom_get( knum, jpdom_autoglo, TRIM(cd_bgcname), p_incs(:,:,:), 1 )
      
      ! Apply the masks
      p_incs(:,:,:) = p_incs(:,:,:) * tmask(:,:,:)
      
      ! Set missing increments to 0.0 rather than 1e+20
      ! to allow for differences in masks
      WHERE( ABS( p_incs(:,:,:) ) > 1.0e+10 ) p_incs(:,:,:) = 0.0

   END SUBROUTINE asm_bgc_read_incs_3d

   !!===========================================================================
   !!===========================================================================
   !!===========================================================================

   SUBROUTINE asm_bgc_init_bkg
      !!------------------------------------------------------------------------
      !!                    ***  ROUTINE asm_bgc_init_bkg  ***
      !!
      !! ** Purpose :   initialise bgc background
      !!
      !! ** Method  :   allocate arrays and read background from file
      !!
      !! ** Action  :   allocate arrays and read background from file
      !!
      !! References :   asm_inc_init
      !!------------------------------------------------------------------------
      !!
      INTEGER :: inum   ! i/o unit of background file
      INTEGER :: jt     ! loop counter
      !!
      !!------------------------------------------------------------------------

#if defined key_top && ( defined key_hadocc || (defined key_medusa && defined key_foam_medusa) )
      IF ( ln_slchltotinc .OR. ln_slchldiainc .OR. ln_slchlnoninc .OR. &
         & ln_schltotinc  .OR. ln_slphytotinc .OR. ln_slphydiainc .OR. &
         & ln_slphynoninc .OR. ln_plchltotinc .OR. ln_pchltotinc ) THEN

         ALLOCATE( pgrow_avg_bkg(jpi,jpj)        )
         ALLOCATE( ploss_avg_bkg(jpi,jpj)        )
         ALLOCATE( phyt_avg_bkg(jpi,jpj)         )
         ALLOCATE( mld_max_bkg(jpi,jpj)          )
         ALLOCATE( tracer_bkg(jpi,jpj,jpk,jptra) )
         pgrow_avg_bkg(:,:)  = 0.0
         ploss_avg_bkg(:,:)  = 0.0
         phyt_avg_bkg(:,:)   = 0.0
         mld_max_bkg(:,:)    = 0.0
         tracer_bkg(:,:,:,:) = 0.0

#if defined key_hadocc
         ALLOCATE( chl_bkg(jpi,jpj,jpk)    )
         ALLOCATE( cchl_p_bkg(jpi,jpj,jpk) )
         chl_bkg(:,:,:)    = 0.0
         cchl_p_bkg(:,:,:) = 0.0
#endif
         
         !--------------------------------------------------------------------
         ! Read background variables for phytoplankton assimilation
         ! Some only required if performing balancing
         !--------------------------------------------------------------------

         CALL iom_open( c_asmbkg, inum )

#if defined key_hadocc
         CALL iom_get( inum, jpdom_autoglo, 'hadocc_chl',  chl_bkg    )
         CALL iom_get( inum, jpdom_autoglo, 'hadocc_cchl', cchl_p_bkg )
         chl_bkg(:,:,:)    = chl_bkg(:,:,:)    * tmask(:,:,:)
         cchl_p_bkg(:,:,:) = cchl_p_bkg(:,:,:) * tmask(:,:,:)
#elif defined key_medusa
         CALL iom_get( inum, jpdom_autoglo, 'medusa_chn', tracer_bkg(:,:,:,jpchn) )
         CALL iom_get( inum, jpdom_autoglo, 'medusa_chd', tracer_bkg(:,:,:,jpchd) )
#endif
         
         IF ( ln_phytobal ) THEN

            CALL iom_get( inum, jpdom_autoglo, 'pgrow_avg', pgrow_avg_bkg )
            CALL iom_get( inum, jpdom_autoglo, 'ploss_avg', ploss_avg_bkg )
            CALL iom_get( inum, jpdom_autoglo, 'phyt_avg',  phyt_avg_bkg  )
            CALL iom_get( inum, jpdom_autoglo, 'mld_max',   mld_max_bkg   )
            pgrow_avg_bkg(:,:) = pgrow_avg_bkg(:,:) * tmask(:,:,1)
            ploss_avg_bkg(:,:) = ploss_avg_bkg(:,:) * tmask(:,:,1)
            phyt_avg_bkg(:,:)  = phyt_avg_bkg(:,:)  * tmask(:,:,1)
            mld_max_bkg(:,:)   = mld_max_bkg(:,:)   * tmask(:,:,1)

#if defined key_hadocc
            CALL iom_get( inum, jpdom_autoglo, 'hadocc_nut', tracer_bkg(:,:,:,jp_had_nut) )
            CALL iom_get( inum, jpdom_autoglo, 'hadocc_phy', tracer_bkg(:,:,:,jp_had_phy) )
            CALL iom_get( inum, jpdom_autoglo, 'hadocc_zoo', tracer_bkg(:,:,:,jp_had_zoo) )
            CALL iom_get( inum, jpdom_autoglo, 'hadocc_pdn', tracer_bkg(:,:,:,jp_had_pdn) )
            CALL iom_get( inum, jpdom_autoglo, 'hadocc_dic', tracer_bkg(:,:,:,jp_had_dic) )
            CALL iom_get( inum, jpdom_autoglo, 'hadocc_alk', tracer_bkg(:,:,:,jp_had_alk) )
#elif defined key_medusa
            CALL iom_get( inum, jpdom_autoglo, 'medusa_phn', tracer_bkg(:,:,:,jpphn) )
            CALL iom_get( inum, jpdom_autoglo, 'medusa_phd', tracer_bkg(:,:,:,jpphd) )
            CALL iom_get( inum, jpdom_autoglo, 'medusa_pds', tracer_bkg(:,:,:,jppds) )
            CALL iom_get( inum, jpdom_autoglo, 'medusa_zmi', tracer_bkg(:,:,:,jpzmi) )
            CALL iom_get( inum, jpdom_autoglo, 'medusa_zme', tracer_bkg(:,:,:,jpzme) )
            CALL iom_get( inum, jpdom_autoglo, 'medusa_din', tracer_bkg(:,:,:,jpdin) )
            CALL iom_get( inum, jpdom_autoglo, 'medusa_sil', tracer_bkg(:,:,:,jpsil) )
            CALL iom_get( inum, jpdom_autoglo, 'medusa_fer', tracer_bkg(:,:,:,jpfer) )
            CALL iom_get( inum, jpdom_autoglo, 'medusa_det', tracer_bkg(:,:,:,jpdet) )
            CALL iom_get( inum, jpdom_autoglo, 'medusa_dtc', tracer_bkg(:,:,:,jpdtc) )
            CALL iom_get( inum, jpdom_autoglo, 'medusa_dic', tracer_bkg(:,:,:,jpdic) )
            CALL iom_get( inum, jpdom_autoglo, 'medusa_alk', tracer_bkg(:,:,:,jpalk) )
            CALL iom_get( inum, jpdom_autoglo, 'medusa_oxy', tracer_bkg(:,:,:,jpoxy) )
#endif
         ELSE IF ( ln_spco2inc .OR. ln_sfco2inc ) THEN
#if defined key_hadocc
            CALL iom_get( inum, jpdom_autoglo, 'hadocc_dic', tracer_bkg(:,:,:,jp_had_dic) )
            CALL iom_get( inum, jpdom_autoglo, 'hadocc_alk', tracer_bkg(:,:,:,jp_had_alk) )
#elif defined key_medusa
            CALL iom_get( inum, jpdom_autoglo, 'medusa_dic', tracer_bkg(:,:,:,jpdic) )
            CALL iom_get( inum, jpdom_autoglo, 'medusa_alk', tracer_bkg(:,:,:,jpalk) )
#endif
            CALL iom_get( inum, jpdom_autoglo, 'mld_max',   mld_max_bkg   )
            mld_max_bkg(:,:) = mld_max_bkg(:,:) * tmask(:,:,1)
         ENDIF

         CALL iom_close( inum )
         
         DO jt = 1, jptra
            tracer_bkg(:,:,:,jt) = tracer_bkg(:,:,:,jt) * tmask(:,:,:)
         END DO
      
      ELSE IF ( ln_spco2inc .OR. ln_sfco2inc .OR. ln_pphinc ) THEN

         ALLOCATE( tracer_bkg(jpi,jpj,jpk,jptra) )
         ALLOCATE( mld_max_bkg(jpi,jpj)          )
         tracer_bkg(:,:,:,:) = 0.0
         mld_max_bkg(:,:)    = 0.0

         CALL iom_open( c_asmbkg, inum )
         
#if defined key_hadocc
         CALL iom_get( inum, jpdom_autoglo, 'hadocc_dic', tracer_bkg(:,:,:,jp_had_dic) )
         CALL iom_get( inum, jpdom_autoglo, 'hadocc_alk', tracer_bkg(:,:,:,jp_had_alk) )
#elif defined key_medusa
         CALL iom_get( inum, jpdom_autoglo, 'medusa_dic', tracer_bkg(:,:,:,jpdic) )
         CALL iom_get( inum, jpdom_autoglo, 'medusa_alk', tracer_bkg(:,:,:,jpalk) )
#endif
         CALL iom_get( inum, jpdom_autoglo, 'mld_max',   mld_max_bkg   )

         CALL iom_close( inum )
         
         DO jt = 1, jptra
            tracer_bkg(:,:,:,jt) = tracer_bkg(:,:,:,jt) * tmask(:,:,:)
         END DO
         mld_max_bkg(:,:) = mld_max_bkg(:,:) * tmask(:,:,1)
      
      ENDIF
#else
      CALL ctl_stop( ' asm_bgc_init_bkg: key_top and a compatible biogeochemical model required' )
#endif

   END SUBROUTINE asm_bgc_init_bkg

   !!===========================================================================
   !!===========================================================================
   !!===========================================================================

   SUBROUTINE asm_bgc_bal_wri( kt )
      !!------------------------------------------------------------------------
      !!
      !!                  ***  ROUTINE asm_bgc_bal_wri ***
      !!
      !! ** Purpose : Write to file the balancing increments
      !!              calculated for biogeochemistry
      !!
      !! ** Method  : Write to file the balancing increments
      !!              calculated for biogeochemistry
      !!
      !! ** Action  :
      !!                   
      !! References :
      !!
      !! History    :
      !!        ! 2014-08 (D. Ford)  Initial version, based on asm_bkg_wri
      !!------------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT( IN ) :: kt        ! Current time-step
      !! * Local declarations
      CHARACTER(LEN=50) :: cl_asmbal     ! Filename (with extension)
      LOGICAL           :: llok          ! Check if file exists
      INTEGER           :: inum          ! File unit number
      REAL(wp)          :: zdate         ! Date
      !!------------------------------------------------------------------------
     
      ! Set things up
      zdate = REAL( ndastp )
      WRITE(cl_asmbal, FMT='(A,".nc")' ) TRIM( c_asmbal )

      ! Check if file exists
      INQUIRE( FILE = TRIM( cl_asmbal ), EXIST = llok )
      IF( .NOT. llok ) THEN
         IF(lwp) WRITE(numout,*) ' Setting up assimilation balancing increments file '// &
            &                    TRIM( c_asmbal ) // ' at timestep = ', kt

         ! Define the output file       
         CALL iom_open( c_asmbal, inum, ldwrt = .TRUE., kiolib = jprstlib)

         ! Write the information
         CALL iom_rstput( kt, kt, inum, 'rdastp' , zdate   )

         IF ( ln_slchltotinc ) THEN
#if defined key_medusa
            CALL iom_rstput( kt, kt, inum, 'logchl_balinc_chn', phyto2d_balinc(:,:,:,jpchn) )
            CALL iom_rstput( kt, kt, inum, 'logchl_balinc_chd', phyto2d_balinc(:,:,:,jpchd) )
            IF ( ln_phytobal ) THEN
               CALL iom_rstput( kt, kt, inum, 'logchl_balinc_phn', phyto2d_balinc(:,:,:,jpphn) )
               CALL iom_rstput( kt, kt, inum, 'logchl_balinc_phd', phyto2d_balinc(:,:,:,jpphd) )
               CALL iom_rstput( kt, kt, inum, 'logchl_balinc_pds', phyto2d_balinc(:,:,:,jppds) )
               CALL iom_rstput( kt, kt, inum, 'logchl_balinc_zmi', phyto2d_balinc(:,:,:,jpzmi) )
               CALL iom_rstput( kt, kt, inum, 'logchl_balinc_zme', phyto2d_balinc(:,:,:,jpzme) )
               CALL iom_rstput( kt, kt, inum, 'logchl_balinc_din', phyto2d_balinc(:,:,:,jpdin) )
               CALL iom_rstput( kt, kt, inum, 'logchl_balinc_sil', phyto2d_balinc(:,:,:,jpsil) )
               CALL iom_rstput( kt, kt, inum, 'logchl_balinc_fer', phyto2d_balinc(:,:,:,jpfer) )
               CALL iom_rstput( kt, kt, inum, 'logchl_balinc_det', phyto2d_balinc(:,:,:,jpdet) )
               CALL iom_rstput( kt, kt, inum, 'logchl_balinc_dtc', phyto2d_balinc(:,:,:,jpdtc) )
               CALL iom_rstput( kt, kt, inum, 'logchl_balinc_dic', phyto2d_balinc(:,:,:,jpdic) )
               CALL iom_rstput( kt, kt, inum, 'logchl_balinc_alk', phyto2d_balinc(:,:,:,jpalk) )
               CALL iom_rstput( kt, kt, inum, 'logchl_balinc_oxy', phyto2d_balinc(:,:,:,jpoxy) )
            ENDIF
#elif defined key_hadocc
            CALL iom_rstput( kt, kt, inum, 'logchl_balinc_phy', phyto2d_balinc(:,:,:,jp_had_phy) )
            IF ( ln_phytobal ) THEN
               CALL iom_rstput( kt, kt, inum, 'logchl_balinc_nut', phyto2d_balinc(:,:,:,jp_had_nut) )
               CALL iom_rstput( kt, kt, inum, 'logchl_balinc_zoo', phyto2d_balinc(:,:,:,jp_had_zoo) )
               CALL iom_rstput( kt, kt, inum, 'logchl_balinc_det', phyto2d_balinc(:,:,:,jp_had_pdn) )
               CALL iom_rstput( kt, kt, inum, 'logchl_balinc_dic', phyto2d_balinc(:,:,:,jp_had_dic) )
               CALL iom_rstput( kt, kt, inum, 'logchl_balinc_alk', phyto2d_balinc(:,:,:,jp_had_alk) )
            ENDIF
#endif
         ENDIF

         IF ( ln_spco2inc ) THEN
#if defined key_medusa
            CALL iom_rstput( kt, kt, inum, 'pco2_balinc_dic', pco2_balinc(:,:,:,jpdic) )
            CALL iom_rstput( kt, kt, inum, 'pco2_balinc_alk', pco2_balinc(:,:,:,jpalk) )
#elif defined key_hadocc
            CALL iom_rstput( kt, kt, inum, 'pco2_balinc_dic', pco2_balinc(:,:,:,jp_had_dic) )
            CALL iom_rstput( kt, kt, inum, 'pco2_balinc_alk', pco2_balinc(:,:,:,jp_had_alk) )
#endif
         ELSE IF ( ln_sfco2inc ) THEN
#if defined key_medusa
            CALL iom_rstput( kt, kt, inum, 'fco2_balinc_dic', pco2_balinc(:,:,:,jpdic) )
            CALL iom_rstput( kt, kt, inum, 'fco2_balinc_alk', pco2_balinc(:,:,:,jpalk) )
#elif defined key_hadocc
            CALL iom_rstput( kt, kt, inum, 'fco2_balinc_dic', pco2_balinc(:,:,:,jp_had_dic) )
            CALL iom_rstput( kt, kt, inum, 'fco2_balinc_alk', pco2_balinc(:,:,:,jp_had_alk) )
#endif
         ENDIF

         CALL iom_close( inum )
      ELSE
         CALL ctl_warn( TRIM( cl_asmbal ) // ' already exists ', &
            &           ' Therefore not writing out balancing increments at this timestep', &
            &           ' Something has probably gone wrong somewhere' )
         IF(lwp) WRITE(numout,*) ' Failed to set up assimilation balancing increments file '// &
            &                    TRIM( c_asmbal ) // ' at timestep = ', kt
      ENDIF
                                 
   END SUBROUTINE asm_bgc_bal_wri

   !!===========================================================================
   !!===========================================================================
   !!===========================================================================

   SUBROUTINE asm_bgc_bkg_wri( kt, knum )
      !!------------------------------------------------------------------------
      !!                    ***  ROUTINE asm_bgc_bkg_wri  ***
      !!
      !! ** Purpose :   write out bgc background
      !!
      !! ** Method  :   write out bgc background
      !!
      !! ** Action  :   write out bgc background
      !!
      !! References :   asm_bkg_wri
      !!------------------------------------------------------------------------
      !!
      INTEGER, INTENT(in   ) :: kt     ! Current time-step
      INTEGER, INTENT(in   ) :: knum   ! i/o unit of increments file
      !!
      !!------------------------------------------------------------------------

#if defined key_hadocc
      CALL iom_rstput( kt, nitbkg_r, knum, 'pgrow_avg'   , pgrow_avg             )
      CALL iom_rstput( kt, nitbkg_r, knum, 'ploss_avg'   , ploss_avg             )
      CALL iom_rstput( kt, nitbkg_r, knum, 'phyt_avg'    , phyt_avg              )
      CALL iom_rstput( kt, nitbkg_r, knum, 'mld_max'     , mld_max               )
      CALL iom_rstput( kt, nitbkg_r, knum, 'hadocc_nut'  , trn(:,:,:,jp_had_nut) )
      CALL iom_rstput( kt, nitbkg_r, knum, 'hadocc_phy'  , trn(:,:,:,jp_had_phy) )
      CALL iom_rstput( kt, nitbkg_r, knum, 'hadocc_zoo'  , trn(:,:,:,jp_had_zoo) )
      CALL iom_rstput( kt, nitbkg_r, knum, 'hadocc_pdn'  , trn(:,:,:,jp_had_pdn) )
      CALL iom_rstput( kt, nitbkg_r, knum, 'hadocc_dic'  , trn(:,:,:,jp_had_dic) )
      CALL iom_rstput( kt, nitbkg_r, knum, 'hadocc_alk'  , trn(:,:,:,jp_had_alk) )
      CALL iom_rstput( kt, nitbkg_r, knum, 'hadocc_chl'  , HADOCC_CHL(:,:,:)     )
      CALL iom_rstput( kt, nitbkg_r, knum, 'hadocc_cchl' , cchl_p(:,:,:)         )
#elif defined key_medusa && defined key_foam_medusa
      CALL iom_rstput( kt, nitbkg_r, knum, 'pgrow_avg'   , pgrow_avg        )
      CALL iom_rstput( kt, nitbkg_r, knum, 'ploss_avg'   , ploss_avg        )
      CALL iom_rstput( kt, nitbkg_r, knum, 'phyt_avg'    , phyt_avg         )
      CALL iom_rstput( kt, nitbkg_r, knum, 'mld_max'     , mld_max          )
      CALL iom_rstput( kt, nitbkg_r, knum, 'medusa_chn'  , trn(:,:,:,jpchn) )
      CALL iom_rstput( kt, nitbkg_r, knum, 'medusa_chd'  , trn(:,:,:,jpchd) )
      CALL iom_rstput( kt, nitbkg_r, knum, 'medusa_phn'  , trn(:,:,:,jpphn) )
      CALL iom_rstput( kt, nitbkg_r, knum, 'medusa_phd'  , trn(:,:,:,jpphd) )
      CALL iom_rstput( kt, nitbkg_r, knum, 'medusa_pds'  , trn(:,:,:,jppds) )
      CALL iom_rstput( kt, nitbkg_r, knum, 'medusa_zmi'  , trn(:,:,:,jpzmi) )
      CALL iom_rstput( kt, nitbkg_r, knum, 'medusa_zme'  , trn(:,:,:,jpzme) )
      CALL iom_rstput( kt, nitbkg_r, knum, 'medusa_din'  , trn(:,:,:,jpdin) )
      CALL iom_rstput( kt, nitbkg_r, knum, 'medusa_sil'  , trn(:,:,:,jpsil) )
      CALL iom_rstput( kt, nitbkg_r, knum, 'medusa_fer'  , trn(:,:,:,jpfer) )
      CALL iom_rstput( kt, nitbkg_r, knum, 'medusa_det'  , trn(:,:,:,jpdet) )
      CALL iom_rstput( kt, nitbkg_r, knum, 'medusa_dtc'  , trn(:,:,:,jpdtc) )
      CALL iom_rstput( kt, nitbkg_r, knum, 'medusa_dic'  , trn(:,:,:,jpdic) )
      CALL iom_rstput( kt, nitbkg_r, knum, 'medusa_alk'  , trn(:,:,:,jpalk) )
      CALL iom_rstput( kt, nitbkg_r, knum, 'medusa_oxy'  , trn(:,:,:,jpoxy) )
#endif

   END SUBROUTINE asm_bgc_bkg_wri

   !!===========================================================================
   !!===========================================================================
   !!===========================================================================

   SUBROUTINE phyto2d_asm_inc( kt, ll_asmdin, ll_asmiau, kcycper, pwgtiau )
      !!------------------------------------------------------------------------
      !!                    ***  ROUTINE phyto2d_asm_inc  ***
      !!          
      !! ** Purpose : Apply the chlorophyll assimilation increments.
      !!
      !! ** Method  : Calculate increments to state variables using nitrogen
      !!              balancing.
      !!              Direct initialization or Incremental Analysis Updating.
      !!
      !! ** Action  : 
      !!------------------------------------------------------------------------
      INTEGER,  INTENT(IN) :: kt        ! Current time step
      LOGICAL,  INTENT(IN) :: ll_asmdin ! Flag for direct initialisation
      LOGICAL,  INTENT(IN) :: ll_asmiau ! Flag for incremental analysis update
      INTEGER,  INTENT(IN) :: kcycper   ! Dimension of pwgtiau
      REAL(wp), DIMENSION(kcycper), INTENT(IN) :: pwgtiau   ! IAU weights
      !
      INTEGER  :: jk              ! Loop counter
      INTEGER  :: it              ! Index
      REAL(wp) :: zincwgt         ! IAU weight for current time step
      REAL(wp) :: zincper         ! IAU interval in seconds
      !!------------------------------------------------------------------------

      IF ( ln_slchldiainc .OR. ln_slchlnoninc .OR. &
         & ln_schltotinc  .OR. ln_slphytotinc .OR. ln_slphydiainc .OR. &
         & ln_slphynoninc ) THEN
         CALL ctl_stop( ' No PFT assimilation quite yet' )
      ENDIF
      
      IF ( kt <= nit000 ) THEN

         zincper = (nitiaufin_r - nitiaustr_r + 1) * rdt

#if defined key_medusa && defined key_foam_medusa
         CALL asm_logchl_bal_medusa( slchltot_bkginc, zincper, mld_choice_bgc, &
            &                        rn_maxchlinc, ln_phytobal, ll_asmdin,  &
            &                        pgrow_avg_bkg, ploss_avg_bkg,           &
            &                        phyt_avg_bkg, mld_max_bkg,              &
            &                        tracer_bkg, phyto2d_balinc )
#elif defined key_hadocc
         CALL asm_logchl_bal_hadocc( slchltot_bkginc, zincper, mld_choice_bgc, &
            &                        rn_maxchlinc, ln_phytobal, ll_asmdin,  &
            &                        pgrow_avg_bkg, ploss_avg_bkg,           &
            &                        phyt_avg_bkg, mld_max_bkg,              &
            &                        chl_bkg(:,:,1), cchl_p_bkg(:,:,1),      &
            &                        tracer_bkg, phyto2d_balinc )
#else
         CALL ctl_stop( 'Attempting to assimilate slchltot, ', &
            &           'but not defined a biogeochemical model' )
#endif

      ENDIF

      IF ( ll_asmiau ) THEN

         !--------------------------------------------------------------------
         ! Incremental Analysis Updating
         !--------------------------------------------------------------------

         IF ( ( kt >= nitiaustr_r ).AND.( kt <= nitiaufin_r ) ) THEN

            it = kt - nit000 + 1
            zincwgt = pwgtiau(it)      ! IAU weight for the current time step
            ! note this is not a tendency so should not be divided by rdt

            IF(lwp) THEN
               WRITE(numout,*) 
               WRITE(numout,*) 'logchl_asm_inc : logchl IAU at time step = ', &
                  &  kt,' with IAU weight = ', pwgtiau(it)
               WRITE(numout,*) '~~~~~~~~~~~~'
            ENDIF

            ! Update the biogeochemical variables
            ! Add directly to trn and trb, rather than to tra, because tra gets
            ! reset to zero at the start of trc_stp, called after this routine
#if defined key_medusa && defined key_foam_medusa
            DO jk = 1, jpkm1
               trn(:,:,jk,jp_msa0:jp_msa1) = trn(:,:,jk,jp_msa0:jp_msa1) + &
                  &                          phyto2d_balinc(:,:,jk,jp_msa0:jp_msa1) * zincwgt
               trb(:,:,jk,jp_msa0:jp_msa1) = trb(:,:,jk,jp_msa0:jp_msa1) + &
                  &                          phyto2d_balinc(:,:,jk,jp_msa0:jp_msa1) * zincwgt
            END DO
#elif defined key_hadocc
            DO jk = 1, jpkm1
               trn(:,:,jk,jp_had0:jp_had1) = trn(:,:,jk,jp_had0:jp_had1) + &
                  &                          phyto2d_balinc(:,:,jk,jp_had0:jp_had1) * zincwgt
               trb(:,:,jk,jp_had0:jp_had1) = trb(:,:,jk,jp_had0:jp_had1) + &
                  &                          phyto2d_balinc(:,:,jk,jp_had0:jp_had1) * zincwgt
            END DO
#endif

            ! Do not deallocate arrays - needed by asm_bgc_bal_wri
            ! which is called at end of model run
         ENDIF

      ELSEIF ( ll_asmdin ) THEN 

         !--------------------------------------------------------------------
         ! Direct Initialization
         !--------------------------------------------------------------------
         
         IF ( kt == nitdin_r ) THEN

            neuler = 0                    ! Force Euler forward step

#if defined key_medusa && defined key_foam_medusa
            ! Initialize the now fields with the background + increment
            ! Background currently is what the model is initialised with
            CALL ctl_warn( ' Doing direct initialisation of MEDUSA with chlorophyll assimilation', &
               &           ' Background state is taken from model rather than background file' )
            trn(:,:,:,jp_msa0:jp_msa1) = trn(:,:,:,jp_msa0:jp_msa1) + &
               &                         phyto2d_balinc(:,:,:,jp_msa0:jp_msa1)
            trb(:,:,:,jp_msa0:jp_msa1) = trn(:,:,:,jp_msa0:jp_msa1)
#elif defined key_hadocc
            ! Initialize the now fields with the background + increment
            ! Background currently is what the model is initialised with
            CALL ctl_warn( ' Doing direct initialisation of HadOCC with chlorophyll assimilation', &
               &           ' Background state is taken from model rather than background file' )
            trn(:,:,:,jp_had0:jp_had1) = trn(:,:,:,jp_had0:jp_had1) + &
               &                         phyto2d_balinc(:,:,:,jp_had0:jp_had1)
            trb(:,:,:,jp_had0:jp_had1) = trn(:,:,:,jp_had0:jp_had1)
#endif
 
            ! Do not deallocate arrays - needed by asm_bgc_bal_wri
            ! which is called at end of model run
         ENDIF
         !
      ENDIF
      !
   END SUBROUTINE phyto2d_asm_inc

   !!===========================================================================
   !!===========================================================================
   !!===========================================================================

   SUBROUTINE phyto3d_asm_inc( kt, ll_asmdin, ll_asmiau, kcycper, pwgtiau )
      !!------------------------------------------------------------------------
      !!                    ***  ROUTINE phyto3d_asm_inc  ***
      !!          
      !! ** Purpose : Apply the profile chlorophyll assimilation increments.
      !!
      !! ** Method  : Calculate increments to state variables.
      !!              Direct initialization or Incremental Analysis Updating.
      !!
      !! ** Action  : 
      !!------------------------------------------------------------------------
      INTEGER,  INTENT(IN) :: kt        ! Current time step
      LOGICAL,  INTENT(IN) :: ll_asmdin ! Flag for direct initialisation
      LOGICAL,  INTENT(IN) :: ll_asmiau ! Flag for incremental analysis update
      INTEGER,  INTENT(IN) :: kcycper   ! Dimension of pwgtiau
      REAL(wp), DIMENSION(kcycper), INTENT(IN) :: pwgtiau   ! IAU weights
      !
      INTEGER                          :: ji, jj, jk ! Loop counters
      INTEGER                          :: it         ! Index
      REAL(wp)                         :: zincwgt    ! IAU weight for time step
      REAL(wp)                         :: zfrac_chn  ! Fraction of jpchn
      REAL(wp)                         :: zfrac_chd  ! Fraction of jpchd
      REAL(wp), DIMENSION(jpi,jpj,jpk) :: chl_inc    ! Chlorophyll increments
      REAL(wp), DIMENSION(jpi,jpj,jpk) :: bkg_chl    ! Chlorophyll background
      !!------------------------------------------------------------------------

      IF ( kt <= nit000 ) THEN

         IF ( ln_plchltotinc ) THEN
            ! Convert log10(chlorophyll) increment back to a chlorophyll increment
            ! In order to transform logchl incs to chl incs, need to account for model
            ! background, cannot simply do 10^logchl_bkginc. Need to:
            ! 1) Add logchl inc to log10(background) to get log10(analysis)
            ! 2) Take 10^log10(analysis) to get analysis
            ! 3) Subtract background from analysis to get chl incs
            ! If rn_maxchlinc > 0 then cap total absolute chlorophyll increment at that value
#if defined key_medusa && defined key_foam_medusa
            bkg_chl(:,:,:) = tracer_bkg(:,:,:,jpchn) + tracer_bkg(:,:,:,jpchd)
#elif defined key_hadocc
            bkg_chl(:,:,:) = chl_bkg(:,:,:)
#endif
            DO jk = 1, jpk
               DO jj = 1, jpj
                  DO ji = 1, jpi
                     IF ( bkg_chl(ji,jj,jk) > 0.0 ) THEN
                        chl_inc(ji,jj,jk) = 10**( LOG10( bkg_chl(ji,jj,jk) ) + plchltot_bkginc(ji,jj,jk) ) - bkg_chl(ji,jj,jk)
                        IF ( rn_maxchlinc > 0.0 ) THEN
                           chl_inc(ji,jj,jk) = MAX( -1.0 * rn_maxchlinc, MIN( chl_inc(ji,jj,jk), rn_maxchlinc ) )
                        ENDIF
                     ELSE
                        chl_inc(ji,jj,jk) = 0.0
                     ENDIF
                  END DO
               END DO
            END DO
         ELSE IF ( ln_pchltotinc ) THEN
            DO jk = 1, jpk
               DO jj = 1, jpj
                  DO ji = 1, jpi
                     IF ( rn_maxchlinc > 0.0 ) THEN
                        chl_inc(ji,jj,jk) = MAX( -1.0 * rn_maxchlinc, MIN( pchltot_bkginc(ji,jj,jk), rn_maxchlinc ) )
                     ELSE
                        chl_inc(ji,jj,jk) = pchltot_bkginc(ji,jj,jk)
                     ENDIF
                  END DO
               END DO
            END DO
         ENDIF

#if defined key_medusa && defined key_foam_medusa
         ! Loop over each grid point partioning the increments based on existing ratios
         DO jk = 1, jpk
            DO jj = 1, jpj
               DO ji = 1, jpi
                  IF ( ( tracer_bkg(ji,jj,jk,jpchn) > 0.0 ) .AND. ( tracer_bkg(ji,jj,jk,jpchd) > 0.0 ) ) THEN
                     zfrac_chn = tracer_bkg(ji,jj,jk,jpchn) / (tracer_bkg(ji,jj,jk,jpchn) + tracer_bkg(ji,jj,jk,jpchd))
                     zfrac_chd = 1.0 - zfrac_chn
                     phyto3d_balinc(ji,jj,jk,jpchn) = chl_inc(ji,jj,jk) * zfrac_chn
                     phyto3d_balinc(ji,jj,jk,jpchd) = chl_inc(ji,jj,jk) * zfrac_chd
                  ENDIF
               END DO
            END DO
         END DO
#elif defined key_hadocc
         phyto3d_balinc(:,:,:,jp_had_phy) = ( cchl_p_bkg(:,:,:) / (mw_carbon * c2n_p) ) * chl_inc(:,:,:)
#else
         CALL ctl_stop( 'Attempting to assimilate p(l)chltot, ', &
            &           'but not defined a biogeochemical model' )
#endif

      ENDIF

      IF ( ll_asmiau ) THEN

         !--------------------------------------------------------------------
         ! Incremental Analysis Updating
         !--------------------------------------------------------------------

         IF ( ( kt >= nitiaustr_r ).AND.( kt <= nitiaufin_r ) ) THEN

            it = kt - nit000 + 1
            zincwgt = pwgtiau(it)      ! IAU weight for the current time step
            ! note this is not a tendency so should not be divided by rdt

            IF(lwp) THEN
               WRITE(numout,*) 
               WRITE(numout,*) 'phyto3d_asm_inc : phyto3d IAU at time step = ', &
                  &  kt,' with IAU weight = ', pwgtiau(it)
               WRITE(numout,*) '~~~~~~~~~~~~'
            ENDIF

            ! Update the biogeochemical variables
            ! Add directly to trn and trb, rather than to tra, because tra gets
            ! reset to zero at the start of trc_stp, called after this routine
#if defined key_medusa && defined key_foam_medusa
            WHERE( phyto3d_balinc(:,:,:,:) > 0.0_wp .OR. &
               &   trn(:,:,:,jp_msa0:jp_msa1) + phyto3d_balinc(:,:,:,:) * zincwgt > 0.0_wp )
               trn(:,:,:,jp_msa0:jp_msa1) = trn(:,:,:,jp_msa0:jp_msa1) + &
                  &                         phyto3d_balinc(:,:,:,jp_msa0:jp_msa1) * zincwgt
               trb(:,:,:,jp_msa0:jp_msa1) = trb(:,:,:,jp_msa0:jp_msa1) + &
                  &                         phyto3d_balinc(:,:,:,jp_msa0:jp_msa1) * zincwgt
            END WHERE
#elif defined key_hadocc
            WHERE( phyto3d_balinc(:,:,:,:) > 0.0_wp .OR. &
               &   trn(:,:,:,jp_had0:jp_had1) + phyto3d_balinc(:,:,:,:) * zincwgt > 0.0_wp )
               trn(:,:,:,jp_had0:jp_had1) = trn(:,:,:,jp_had0:jp_had1) + &
                  &                         phyto3d_balinc(:,:,:,jp_had0:jp_had1) * zincwgt
               trb(:,:,:,jp_had0:jp_had1) = trb(:,:,:,jp_had0:jp_had1) + &
                  &                         phyto3d_balinc(:,:,:,jp_had0:jp_had1) * zincwgt
            END WHERE
#endif

            ! Do not deallocate arrays - needed by asm_bgc_bal_wri
            ! which is called at end of model run
         ENDIF

      ELSEIF ( ll_asmdin ) THEN 

         !--------------------------------------------------------------------
         ! Direct Initialization
         !--------------------------------------------------------------------
         
         IF ( kt == nitdin_r ) THEN

            neuler = 0                    ! Force Euler forward step

            ! Initialize the now fields with the background + increment
            ! Background currently is what the model is initialised with
            CALL ctl_warn( ' Doing direct initialisation with phyto3d assimilation', &
               &           ' Background state is taken from model rather than background file' )
#if defined key_medusa && defined key_foam_medusa
            WHERE( phyto3d_balinc(:,:,:,:) > 0.0_wp .OR. &
               &   trn(:,:,:,jp_msa0:jp_msa1) + phyto3d_balinc(:,:,:,:) * zincwgt > 0.0_wp )
               trn(:,:,:,jp_msa0:jp_msa1) = trn(:,:,:,jp_msa0:jp_msa1) + &
                  &                         phyto3d_balinc(:,:,:,jp_msa0:jp_msa1)
               trb(:,:,:,jp_msa0:jp_msa1) = trn(:,:,:,jp_msa0:jp_msa1)
            END WHERE
#elif defined key_hadocc
            WHERE( phyto3d_balinc(:,:,:,:) > 0.0_wp .OR. &
               &   trn(:,:,:,jp_had0:jp_had1) + phyto3d_balinc(:,:,:,:) * zincwgt > 0.0_wp )
               trn(:,:,:,jp_had0:jp_had1) = trn(:,:,:,jp_had0:jp_had1) + &
                  &                         phyto3d_balinc(:,:,:,jp_had0:jp_had1)
               trb(:,:,:,jp_had0:jp_had1) = trn(:,:,:,jp_had0:jp_had1)
            END WHERE
#endif
 
            ! Do not deallocate arrays - needed by asm_bgc_bal_wri
            ! which is called at end of model run
         ENDIF
         !
      ENDIF
      !
   END SUBROUTINE phyto3d_asm_inc

   !!===========================================================================
   !!===========================================================================
   !!===========================================================================

   SUBROUTINE pco2_asm_inc( kt, ll_asmdin, ll_asmiau, kcycper, pwgtiau, &
      &                     ll_trainc, pt_bkginc, ps_bkginc )
      !!------------------------------------------------------------------------
      !!                    ***  ROUTINE pco2_asm_inc  ***
      !!          
      !! ** Purpose : Apply the pco2/fco2 assimilation increments.
      !!
      !! ** Method  : Calculate increments to state variables using carbon
      !!              balancing.
      !!              Direct initialization or Incremental Analysis Updating.
      !!
      !! ** Action  : 
      !!------------------------------------------------------------------------
      INTEGER, INTENT(IN)                   :: kt           ! Current time step
      LOGICAL, INTENT(IN)                   :: ll_asmdin    ! Flag for direct initialisation
      LOGICAL, INTENT(IN)                   :: ll_asmiau    ! Flag for incremental analysis update
      INTEGER, INTENT(IN)                   :: kcycper      ! Dimension of pwgtiau
      REAL(wp), DIMENSION(kcycper), INTENT(IN) :: pwgtiau   ! IAU weights
      LOGICAL, INTENT(IN)                   :: ll_trainc    ! Flag for T/S increments
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT(IN) :: pt_bkginc ! T increments
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT(IN) :: ps_bkginc ! S increments
      !
      INTEGER                               :: ji, jj, jk   ! Loop counters
      INTEGER                               :: it           ! Index
      INTEGER                               :: jkmax        ! Index of mixed layer depth
      REAL(wp)                              :: zincwgt      ! IAU weight for current time step
      REAL(wp), DIMENSION(jpi,jpj)          :: zmld         ! Mixed layer depth for balancing
      REAL(wp), DIMENSION(jpi,jpj)          :: pco2_bkginc_temp ! For fCO2 conversion
      REAL(wp), DIMENSION(jpi,jpj)          :: dic_bkg_temp ! DIC background
      REAL(wp), DIMENSION(jpi,jpj)          :: alk_bkg_temp ! Alkalinity background
      REAL(wp), DIMENSION(jpi,jpj)          :: tem_bkg_temp ! Temperature background
      REAL(wp), DIMENSION(jpi,jpj)          :: sal_bkg_temp ! Salinity background
      REAL(wp), DIMENSION(1)                :: patm         ! Atmospheric pressure
      REAL(wp), DIMENSION(1)                :: phyd         ! Hydrostatic pressure

      ! Coefficients for fCO2 to pCO2 conversion
      REAL(wp)                              :: zcoef_fco2_1 = -1636.75
      REAL(wp)                              :: zcoef_fco2_2 = 12.0408
      REAL(wp)                              :: zcoef_fco2_3 = 0.0327957
      REAL(wp)                              :: zcoef_fco2_4 = 0.0000316528
      REAL(wp)                              :: zcoef_fco2_5 = 2.0
      REAL(wp)                              :: zcoef_fco2_6 = 57.7
      REAL(wp)                              :: zcoef_fco2_7 = 0.118
      REAL(wp)                              :: zcoef_fco2_8 = 82.0578
      !!------------------------------------------------------------------------

      IF ( kt <= nit000 ) THEN

         !----------------------------------------------------------------------
         ! DIC and alkalinity balancing
         !----------------------------------------------------------------------

         IF ( ln_sfco2inc ) THEN
#if defined key_medusa && defined key_roam
            ! If assimilating fCO2, then convert to pCO2 using MEDUSA MOCSY subroutine
            patm(1) = 1.0
            phyd(1) = 0.0
            DO jj = 1, jpj
               DO ji = 1, jpi
                  CALL f2pCO2( sfco2_bkginc(ji,jj), tsn(ji,jj,1,1), patm, phyd, 1, pco2_bkginc_temp(ji,jj) )
               END DO
            END DO
#else
            ! If assimilating fCO2, then convert to pCO2 using temperature
            ! See flux_gas.F90 within HadOCC for details of calculation
            pco2_bkginc_temp(:,:) = sfco2_bkginc(:,:) /                                                            &
               &                    EXP((zcoef_fco2_1                                                            + &
               &                         zcoef_fco2_2 * (tsn(:,:,1,1)+rt0)                                       - &
               &                         zcoef_fco2_3 * (tsn(:,:,1,1)+rt0)*(tsn(:,:,1,1)+rt0)                    + &
               &                         zcoef_fco2_4 * (tsn(:,:,1,1)+rt0)*(tsn(:,:,1,1)+rt0)*(tsn(:,:,1,1)+rt0) + &
               &                         zcoef_fco2_5 * (zcoef_fco2_6 - zcoef_fco2_7 * (tsn(:,:,1,1)+rt0)))      / &
               &                        (zcoef_fco2_8 * (tsn(:,:,1,1)+rt0)))
#endif
         ELSE
            pco2_bkginc_temp(:,:) = spco2_bkginc(:,:)
         ENDIF

         ! Account for physics assimilation if required
         IF ( ll_trainc ) THEN
            tem_bkg_temp(:,:) = tsn(:,:,1,1) + pt_bkginc(:,:,1)
            sal_bkg_temp(:,:) = tsn(:,:,1,2) + ps_bkginc(:,:,1)
         ELSE
            tem_bkg_temp(:,:) = tsn(:,:,1,1)
            sal_bkg_temp(:,:) = tsn(:,:,1,2)
         ENDIF

#if defined key_medusa
         ! Account for logchl balancing if required
         IF ( ln_slchltotinc .AND. ln_phytobal ) THEN
            dic_bkg_temp(:,:) = tracer_bkg(:,:,1,jpdic) + phyto2d_balinc(:,:,1,jpdic)
            alk_bkg_temp(:,:) = tracer_bkg(:,:,1,jpalk) + phyto2d_balinc(:,:,1,jpalk)
         ELSE
            dic_bkg_temp(:,:) = tracer_bkg(:,:,1,jpdic)
            alk_bkg_temp(:,:) = tracer_bkg(:,:,1,jpalk)
         ENDIF

         CALL asm_pco2_bal( pco2_bkginc_temp(:,:), dic_bkg_temp(:,:), alk_bkg_temp(:,:), &
            &               tem_bkg_temp(:,:), sal_bkg_temp(:,:),                        &
            &               pco2_balinc(:,:,1,jpdic), pco2_balinc(:,:,1,jpalk) )

#elif defined key_hadocc
         ! Account for slchltot balancing if required
         IF ( ln_slchltotinc .AND. ln_phytobal ) THEN
            dic_bkg_temp(:,:) = tracer_bkg(:,:,1,jp_had_dic) + phyto2d_balinc(:,:,1,jp_had_dic)
            alk_bkg_temp(:,:) = tracer_bkg(:,:,1,jp_had_alk) + phyto2d_balinc(:,:,1,jp_had_alk)
         ELSE
            dic_bkg_temp(:,:) = tracer_bkg(:,:,1,jp_had_dic)
            alk_bkg_temp(:,:) = tracer_bkg(:,:,1,jp_had_alk)
         ENDIF

         CALL asm_pco2_bal( pco2_bkginc_temp(:,:), dic_bkg_temp(:,:), alk_bkg_temp(:,:), &
            &               tem_bkg_temp(:,:), sal_bkg_temp(:,:),                        &
            &               pco2_balinc(:,:,1,jp_had_dic), pco2_balinc(:,:,1,jp_had_alk) )

#else
         CALL ctl_stop( 'Attempting to assimilate pCO2/fCO2, ', &
            &           'but not defined a biogeochemical model' )
#endif

         ! Select mixed layer
         IF ( ll_asmdin ) THEN
#if defined key_hadocc || (defined key_medusa && defined key_foam_medusa)
            CALL ctl_warn( ' Doing direct initialisation with pCO2 assimilation', &
               &           ' Mixed layer depth taken to be background maximum mld_max_bkg' )
            zmld(:,:) = mld_max_bkg(:,:)
#else
            CALL ctl_stop( 'No biogeochemical model defined for pCO2 assimilation' )
#endif
         ELSE
            SELECT CASE( mld_choice_bgc )
            CASE ( 1 )                   ! Turbocline/mixing depth [W points]
               zmld(:,:) = hmld(:,:)
            CASE ( 2 )                   ! Density criterion (0.01 kg/m^3 change from 10m) [W points]
               zmld(:,:) = hmlp(:,:)
            CASE ( 3 )                   ! Kara MLD [Interpolated]
#if defined key_karaml
               IF ( ln_kara ) THEN
                  zmld(:,:) = hmld_kara(:,:)
               ELSE
                  CALL ctl_stop( ' Kara mixed layer requested for pCO2 assimilation,', &
                     &           ' but ln_kara=.false.' )
               ENDIF
#else
               CALL ctl_stop( ' Kara mixed layer requested for pCO2 assimilation,', &
                  &           ' but is not defined' )
#endif
            CASE ( 4 )                   ! Temperature criterion (0.2 K change from surface) [T points]
               !zmld(:,:) = hmld_tref(:,:)
               CALL ctl_stop( ' hmld_tref mixed layer requested for pCO2 assimilation,', &
                  &           ' but is not available in this version' )
            CASE ( 5 )                   ! Density criterion (0.01 kg/m^3 change from 10m) [T points]
               zmld(:,:) = hmlpt(:,:)
            END SELECT
         ENDIF

         ! Propagate through mixed layer
         DO jj = 1, jpj
            DO ji = 1, jpi
               !
               jkmax = jpk-1
               DO jk = jpk-1, 1, -1
                  IF ( ( zmld(ji,jj) >  gdepw_n(ji,jj,jk)   ) .AND. &
                     & ( zmld(ji,jj) <= gdepw_n(ji,jj,jk+1) ) ) THEN
                     zmld(ji,jj) = gdepw_n(ji,jj,jk+1)
                     jkmax = jk
                  ENDIF
               END DO
               !
#if defined key_top
               DO jk = 2, jkmax
                  pco2_balinc(ji,jj,jk,:) = pco2_balinc(ji,jj,1,:)
               END DO
#endif
               !
            END DO
         END DO

      ENDIF

      IF ( ll_asmiau ) THEN

         !--------------------------------------------------------------------
         ! Incremental Analysis Updating
         !--------------------------------------------------------------------

         IF ( ( kt >= nitiaustr_r ).AND.( kt <= nitiaufin_r ) ) THEN

            it = kt - nit000 + 1
            zincwgt = pwgtiau(it)      ! IAU weight for the current time step
            ! note this is not a tendency so should not be divided by rdt

            IF(lwp) THEN
               WRITE(numout,*) 
               IF ( ln_spco2inc ) THEN
                  WRITE(numout,*) 'pco2_asm_inc : pco2 IAU at time step = ', &
                     &  kt,' with IAU weight = ', pwgtiau(it)
               ELSE IF ( ln_sfco2inc ) THEN
                  WRITE(numout,*) 'fco2_asm_inc : fco2 IAU at time step = ', &
                     &  kt,' with IAU weight = ', pwgtiau(it)
               ENDIF
               WRITE(numout,*) '~~~~~~~~~~~~'
            ENDIF

            ! Update the biogeochemical variables
            ! Add directly to trn and trb, rather than to tra, because tra gets
            ! reset to zero at the start of trc_stp, called after this routine
#if defined key_medusa && defined key_foam_medusa
            DO jk = 1, jpkm1
               trn(:,:,jk,jp_msa0:jp_msa1) = trn(:,:,jk,jp_msa0:jp_msa1) + &
                  &                          pco2_balinc(:,:,jk,jp_msa0:jp_msa1) * zincwgt
               trb(:,:,jk,jp_msa0:jp_msa1) = trb(:,:,jk,jp_msa0:jp_msa1) + &
                  &                          pco2_balinc(:,:,jk,jp_msa0:jp_msa1) * zincwgt
            END DO
#elif defined key_hadocc
            DO jk = 1, jpkm1
               trn(:,:,jk,jp_had0:jp_had1) = trn(:,:,jk,jp_had0:jp_had1) + &
                  &                          pco2_balinc(:,:,jk,jp_had0:jp_had1) * zincwgt
               trb(:,:,jk,jp_had0:jp_had1) = trb(:,:,jk,jp_had0:jp_had1) + &
                  &                          pco2_balinc(:,:,jk,jp_had0:jp_had1) * zincwgt
            END DO
#endif

            ! Do not deallocate arrays - needed by asm_bgc_bal_wri
            ! which is called at end of model run

         ENDIF

      ELSEIF ( ll_asmdin ) THEN 

         !--------------------------------------------------------------------
         ! Direct Initialization
         !--------------------------------------------------------------------
         
         IF ( kt == nitdin_r ) THEN

            neuler = 0                    ! Force Euler forward step

#if defined key_medusa && defined key_foam_medusa
            ! Initialize the now fields with the background + increment
            ! Background currently is what the model is initialised with
            CALL ctl_warn( ' Doing direct initialisation of MEDUSA with pCO2 assimilation', &
               &           ' Background state is taken from model rather than background file' )
            trn(:,:,:,jp_msa0:jp_msa1) = trn(:,:,:,jp_msa0:jp_msa1) + &
               &                         pco2_balinc(:,:,:,jp_msa0:jp_msa1)
            trb(:,:,:,jp_msa0:jp_msa1) = trn(:,:,:,jp_msa0:jp_msa1)
#elif defined key_hadocc
            ! Initialize the now fields with the background + increment
            ! Background currently is what the model is initialised with
            CALL ctl_warn( ' Doing direct initialisation of HadOCC with pCO2 assimilation', &
               &           ' Background state is taken from model rather than background file' )
            trn(:,:,:,jp_had0:jp_had1) = trn(:,:,:,jp_had0:jp_had1) + &
               &                         pco2_balinc(:,:,:,jp_had0:jp_had1)
            trb(:,:,:,jp_had0:jp_had1) = trn(:,:,:,jp_had0:jp_had1)
#endif
 
            ! Do not deallocate arrays - needed by asm_bgc_bal_wri
            ! which is called at end of model run
         ENDIF
         !
      ENDIF
      !
   END SUBROUTINE pco2_asm_inc

   !!===========================================================================
   !!===========================================================================
   !!===========================================================================

   SUBROUTINE ph_asm_inc( kt, ll_asmdin, ll_asmiau, kcycper, pwgtiau, &
      &                   ll_trainc, pt_bkginc, ps_bkginc )
      !!------------------------------------------------------------------------
      !!                    ***  ROUTINE ph_asm_inc  ***
      !!          
      !! ** Purpose : Apply the pH assimilation increments.
      !!
      !! ** Method  : Calculate increments to state variables using carbon
      !!              balancing.
      !!              Direct initialization or Incremental Analysis Updating.
      !!
      !! ** Action  : 
      !!------------------------------------------------------------------------
      INTEGER, INTENT(IN)                   :: kt           ! Current time step
      LOGICAL, INTENT(IN)                   :: ll_asmdin    ! Flag for direct initialisation
      LOGICAL, INTENT(IN)                   :: ll_asmiau    ! Flag for incremental analysis update
      INTEGER, INTENT(IN)                   :: kcycper      ! Dimension of pwgtiau
      REAL(wp), DIMENSION(kcycper), INTENT(IN) :: pwgtiau   ! IAU weights
      LOGICAL, INTENT(IN)                   :: ll_trainc    ! Flag for T/S increments
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT(IN) :: pt_bkginc ! T increments
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT(IN) :: ps_bkginc ! S increments
      !!------------------------------------------------------------------------
      
      CALL ctl_stop( ' pH balancing not yet implemented' )
      
      ! See solve_at_general line 281 of mocsy_phsolvers.F90
      !
      ! Or, call to mocsy_interface line 158 of carb_chem.F90
      ! Input variables (rest are output) are:
      ! ztmp(ji,jj),zsal(ji,jj),zalk(ji,jj), &
      ! zdic(ji,jj),zsil(ji,jj),zpho(ji,jj), &
      ! f_pp0(ji,jj),fsdept(ji,jj,jk),       &
      ! gphit(ji,jj),f_kw660(ji,jj),         &
      ! f_xco2a(ji,jj),1
      !
      ! ztmp    = tsn(:,:,:,jp_tem)
      ! zsal    = tsn(:,:,:,jp_sal)
      ! zalk    = trn(:,:,:,jpalk)
      ! zdic    = trn(:,:,:,jpdic)
      ! zsil    = trn(:,:,:,jpsil)
      ! zpho    = trn(:,:,:,jpdin) / 16.0
      ! f_pp0   = 1.0
      ! fsdept  = fsdept(:,:,:)
      ! gphit   = gphit(:,:)
      ! f_kw660 = 1.0
      ! f_xco2a = f_xco2a(:,:)
      
      !
   END SUBROUTINE ph_asm_inc

   !!===========================================================================
   !!===========================================================================
   !!===========================================================================

   SUBROUTINE bgc3d_asm_inc( kt, ll_asmdin, ll_asmiau, kcycper, pwgtiau )
      !!----------------------------------------------------------------------
      !!                    ***  ROUTINE dyn_asm_inc  ***
      !!          
      !! ** Purpose : Apply generic 3D biogeochemistry assimilation increments.
      !!
      !! ** Method  : Direct initialization or Incremental Analysis Updating.
      !!
      !! ** Action  : 
      !!----------------------------------------------------------------------
      INTEGER,  INTENT(IN) :: kt        ! Current time step
      LOGICAL,  INTENT(IN) :: ll_asmdin ! Flag for direct initialisation
      LOGICAL,  INTENT(IN) :: ll_asmiau ! Flag for incremental analysis update
      INTEGER,  INTENT(IN) :: kcycper   ! Dimension of pwgtiau
      REAL(wp), DIMENSION(kcycper), INTENT(IN) :: pwgtiau   ! IAU weights
      !
      INTEGER  :: jk              ! Loop counter
      INTEGER  :: it              ! Index
      REAL(wp) :: zincwgt         ! IAU weight for current time step
      REAL(wp) :: zincper         ! IAU interval in seconds
      !!----------------------------------------------------------------------

      IF ( ll_asmiau ) THEN

         !--------------------------------------------------------------------
         ! Incremental Analysis Updating
         !--------------------------------------------------------------------

         IF ( ( kt >= nitiaustr_r ).AND.( kt <= nitiaufin_r ) ) THEN

            it = kt - nit000 + 1
            zincwgt = pwgtiau(it)      ! IAU weight for the current time step
            ! note this is not a tendency so should not be divided by rdt

            IF(lwp) THEN
               WRITE(numout,*) 
               WRITE(numout,*) 'bgc3d_asm_inc : 3D BGC IAU at time step = ', &
                  &  kt,' with IAU weight = ', pwgtiau(it)
               WRITE(numout,*) '~~~~~~~~~~~~'
            ENDIF

            ! Update the 3D BGC variables
            ! Add directly to trn and trb, rather than to tra, because tra gets
            ! reset to zero at the start of trc_stp, called after this routine
            ! Don't apply increments if they'll take concentrations negative

            IF ( ln_pno3inc ) THEN
#if defined key_hadocc
               WHERE( pno3_bkginc(:,:,:) > 0.0_wp .OR. &
                  &   trn(:,:,:,jp_had_nut) + pno3_bkginc(:,:,:) * zincwgt > 0.0_wp )
                  trn(:,:,:,jp_had_nut) = trn(:,:,:,jp_had_nut) + pno3_bkginc(:,:,:) * zincwgt
                  trb(:,:,:,jp_had_nut) = trb(:,:,:,jp_had_nut) + pno3_bkginc(:,:,:) * zincwgt
               END WHERE
#elif defined key_medusa && defined key_foam_medusa
               WHERE( pno3_bkginc(:,:,:) > 0.0_wp .OR. &
                  &   trn(:,:,:,jpdin) + pno3_bkginc(:,:,:) * zincwgt > 0.0_wp )
                  trn(:,:,:,jpdin) = trn(:,:,:,jpdin) + pno3_bkginc(:,:,:) * zincwgt
                  trb(:,:,:,jpdin) = trb(:,:,:,jpdin) + pno3_bkginc(:,:,:) * zincwgt
               END WHERE
#else
               CALL ctl_stop ( ' bgc3d_asm_inc: no compatible BGC model defined' )
#endif
            ENDIF

            IF ( ln_psi4inc ) THEN
#if defined key_medusa && defined key_foam_medusa
               WHERE( psi4_bkginc(:,:,:) > 0.0_wp .OR. &
                  &   trn(:,:,:,jpsil) + psi4_bkginc(:,:,:) * zincwgt > 0.0_wp )
                  trn(:,:,:,jpsil) = trn(:,:,:,jpsil) + psi4_bkginc(:,:,:) * zincwgt
                  trb(:,:,:,jpsil) = trb(:,:,:,jpsil) + psi4_bkginc(:,:,:) * zincwgt
               END WHERE
#else
               CALL ctl_stop ( ' bgc3d_asm_inc: no compatible BGC model defined' )
#endif
            ENDIF

            IF ( ln_pdicinc ) THEN
#if defined key_hadocc
               WHERE( pdic_bkginc(:,:,:) > 0.0_wp .OR. &
                  &   trn(:,:,:,jp_had_dic) + pdic_bkginc(:,:,:) * zincwgt > 0.0_wp )
                  trn(:,:,:,jp_had_dic) = trn(:,:,:,jp_had_dic) + pdic_bkginc(:,:,:) * zincwgt
                  trb(:,:,:,jp_had_dic) = trb(:,:,:,jp_had_dic) + pdic_bkginc(:,:,:) * zincwgt
               END WHERE
#elif defined key_medusa && defined key_foam_medusa
               WHERE( pdic_bkginc(:,:,:) > 0.0_wp .OR. &
                  &   trn(:,:,:,jpdic) + pdic_bkginc(:,:,:) * zincwgt > 0.0_wp )
                  trn(:,:,:,jpdic) = trn(:,:,:,jpdic) + pdic_bkginc(:,:,:) * zincwgt
                  trb(:,:,:,jpdic) = trb(:,:,:,jpdic) + pdic_bkginc(:,:,:) * zincwgt
               END WHERE
#else
               CALL ctl_stop ( ' bgc3d_asm_inc: no compatible BGC model defined' )
#endif
            ENDIF

            IF ( ln_palkinc ) THEN
#if defined key_hadocc
               WHERE( palk_bkginc(:,:,:) > 0.0_wp .OR. &
                  &   trn(:,:,:,jp_had_alk) + palk_bkginc(:,:,:) * zincwgt > 0.0_wp )
                  trn(:,:,:,jp_had_alk) = trn(:,:,:,jp_had_alk) + palk_bkginc(:,:,:) * zincwgt
                  trb(:,:,:,jp_had_alk) = trb(:,:,:,jp_had_alk) + palk_bkginc(:,:,:) * zincwgt
               END WHERE
#elif defined key_medusa && defined key_foam_medusa
               WHERE( palk_bkginc(:,:,:) > 0.0_wp .OR. &
                  &   trn(:,:,:,jpalk) + palk_bkginc(:,:,:) * zincwgt > 0.0_wp )
                  trn(:,:,:,jpalk) = trn(:,:,:,jpalk) + palk_bkginc(:,:,:) * zincwgt
                  trb(:,:,:,jpalk) = trb(:,:,:,jpalk) + palk_bkginc(:,:,:) * zincwgt
               END WHERE
#else
               CALL ctl_stop ( ' bgc3d_asm_inc: no compatible BGC model defined' )
#endif
            ENDIF

            IF ( ln_po2inc ) THEN
#if defined key_medusa && defined key_foam_medusa
               WHERE( po2_bkginc(:,:,:) > 0.0_wp .OR. &
                  &   trn(:,:,:,jpoxy) + po2_bkginc(:,:,:) * zincwgt > 0.0_wp )
                  trn(:,:,:,jpoxy) = trn(:,:,:,jpoxy) + po2_bkginc(:,:,:) * zincwgt
                  trb(:,:,:,jpoxy) = trb(:,:,:,jpoxy) + po2_bkginc(:,:,:) * zincwgt
               END WHERE
#else
               CALL ctl_stop ( ' bgc3d_asm_inc: no compatible BGC model defined' )
#endif
            ENDIF
           
            IF ( kt == nitiaufin_r ) THEN
               IF ( ln_pno3inc ) DEALLOCATE( pno3_bkginc )
               IF ( ln_psi4inc ) DEALLOCATE( psi4_bkginc )
               IF ( ln_pdicinc ) DEALLOCATE( pdic_bkginc )
               IF ( ln_palkinc ) DEALLOCATE( palk_bkginc )
               IF ( ln_po2inc  ) DEALLOCATE( po2_bkginc  )
            ENDIF

         ENDIF

      ELSEIF ( ll_asmdin ) THEN 

         !--------------------------------------------------------------------
         ! Direct Initialization
         !--------------------------------------------------------------------
         
         IF ( kt == nitdin_r ) THEN

            neuler = 0                    ! Force Euler forward step

            ! Initialize the now fields with the background + increment
            ! Background currently is what the model is initialised with
#if defined key_hadocc
            CALL ctl_warn( ' Doing direct initialisation of HadOCC with 3D BGC assimilation', &
               &           ' Background state is taken from model rather than background file' )
#elif defined key_medusa && defined key_foam_medusa
            CALL ctl_warn( ' Doing direct initialisation of MEDUSA with 3D BGC assimilation', &
               &           ' Background state is taken from model rather than background file' )
#endif

            trn(:,:,:,jp_msa0:jp_msa1) = trn(:,:,:,jp_msa0:jp_msa1) + &
               &                         pco2_balinc(:,:,:,jp_msa0:jp_msa1)
            trb(:,:,:,jp_msa0:jp_msa1) = trn(:,:,:,jp_msa0:jp_msa1)

            IF ( ln_pno3inc ) THEN
#if defined key_hadocc
               WHERE( pno3_bkginc(:,:,:) > 0.0_wp .OR. &
                  &   trn(:,:,:,jp_had_nut) + pno3_bkginc(:,:,:) > 0.0_wp )
                  trn(:,:,:,jp_had_nut) = trn(:,:,:,jp_had_nut) + pno3_bkginc(:,:,:)
                  trb(:,:,:,jp_had_nut) = trn(:,:,:,jp_had_nut)
               END WHERE
#elif defined key_medusa && defined key_foam_medusa
               WHERE( pno3_bkginc(:,:,:) > 0.0_wp .OR. &
                  &   trn(:,:,:,jpdin) + pno3_bkginc(:,:,:) > 0.0_wp )
                  trn(:,:,:,jpdin) = trn(:,:,:,jpdin) + pno3_bkginc(:,:,:)
                  trb(:,:,:,jpdin) = trn(:,:,:,jpdin)
               END WHERE
#else
               CALL ctl_stop ( ' bgc3d_asm_inc: no compatible BGC model defined' )
#endif
            ENDIF

            IF ( ln_psi4inc ) THEN
#if defined key_medusa && defined key_foam_medusa
               WHERE( psi4_bkginc(:,:,:) > 0.0_wp .OR. &
                  &   trn(:,:,:,jpsil) + psi4_bkginc(:,:,:) > 0.0_wp )
                  trn(:,:,:,jpsil) = trn(:,:,:,jpsil) + psi4_bkginc(:,:,:)
                  trb(:,:,:,jpsil) = trn(:,:,:,jpsil)
               END WHERE
#else
               CALL ctl_stop ( ' bgc3d_asm_inc: no compatible BGC model defined' )
#endif
            ENDIF

            IF ( ln_pdicinc ) THEN
#if defined key_hadocc
               WHERE( pdic_bkginc(:,:,:) > 0.0_wp .OR. &
                  &   trn(:,:,:,jp_had_dic) + pdic_bkginc(:,:,:) > 0.0_wp )
                  trn(:,:,:,jp_had_dic) = trn(:,:,:,jp_had_dic) + pdic_bkginc(:,:,:)
                  trb(:,:,:,jp_had_dic) = trn(:,:,:,jp_had_dic)
               END WHERE
#elif defined key_medusa && defined key_foam_medusa
               WHERE( pdic_bkginc(:,:,:) > 0.0_wp .OR. &
                  &   trn(:,:,:,jpdic) + pdic_bkginc(:,:,:) > 0.0_wp )
                  trn(:,:,:,jpdic) = trn(:,:,:,jpdic) + pdic_bkginc(:,:,:)
                  trb(:,:,:,jpdic) = trn(:,:,:,jpdic)
               END WHERE
#else
               CALL ctl_stop ( ' bgc3d_asm_inc: no compatible BGC model defined' )
#endif
            ENDIF

            IF ( ln_palkinc ) THEN
#if defined key_hadocc
               WHERE( palk_bkginc(:,:,:) > 0.0_wp .OR. &
                  &   trn(:,:,:,jp_had_alk) + palk_bkginc(:,:,:) > 0.0_wp )
                  trn(:,:,:,jp_had_alk) = trn(:,:,:,jp_had_alk) + palk_bkginc(:,:,:)
                  trb(:,:,:,jp_had_alk) = trn(:,:,:,jp_had_alk)
               END WHERE
#elif defined key_medusa && defined key_foam_medusa
               WHERE( palk_bkginc(:,:,:) > 0.0_wp .OR. &
                  &   trn(:,:,:,jpalk) + palk_bkginc(:,:,:) > 0.0_wp )
                  trn(:,:,:,jpalk) = trn(:,:,:,jpalk) + palk_bkginc(:,:,:)
                  trb(:,:,:,jpalk) = trn(:,:,:,jpalk)
               END WHERE
#else
               CALL ctl_stop ( ' bgc3d_asm_inc: no compatible BGC model defined' )
#endif
            ENDIF

            IF ( ln_po2inc ) THEN
#if defined key_medusa && defined key_foam_medusa
               WHERE( po2_bkginc(:,:,:) > 0.0_wp .OR. &
                  &   trn(:,:,:,jpoxy) + po2_bkginc(:,:,:) > 0.0_wp )
                  trn(:,:,:,jpoxy) = trn(:,:,:,jpoxy) + po2_bkginc(:,:,:)
                  trb(:,:,:,jpoxy) = trn(:,:,:,jpoxy)
               END WHERE
#else
               CALL ctl_stop ( ' bgc3d_asm_inc: no compatible BGC model defined' )
#endif
            ENDIF
 
            IF ( ln_pno3inc ) DEALLOCATE( pno3_bkginc )
            IF ( ln_psi4inc ) DEALLOCATE( psi4_bkginc )
            IF ( ln_pdicinc ) DEALLOCATE( pdic_bkginc )
            IF ( ln_palkinc ) DEALLOCATE( palk_bkginc )
            IF ( ln_po2inc  ) DEALLOCATE( po2_bkginc  )
         ENDIF
         !
      ENDIF
      !
   END SUBROUTINE bgc3d_asm_inc

   !!===========================================================================

END MODULE asmbgc
